@@ -12,9 +12,10 @@ from cyclone.web import HTTPError
 from go_api.collections import InMemoryCollection
 from go_api.collections.errors import CollectionUsageError
 from go_api.cyclone.handlers import (
-    BaseHandler, CollectionHandler, ElementHandler,
-    create_urlspec_regex, ApiApplication,
-    owner_from_header, owner_from_path_kwarg, owner_from_oauth2_bouncer)
+    RouteParseError, BaseHandler, CollectionHandler, ElementHandler,
+    duplicates, join_paths, parse_route_vars, create_urlspec_regex,
+    ApiApplication, owner_from_header, owner_from_path_kwarg,
+    owner_from_oauth2_bouncer)
 from go_api.cyclone.helpers import HandlerHelper, AppHelper, MockHttpServer
 
 
@@ -24,6 +25,11 @@ class DummyError(Exception):
     """
 
 
+class DummyHandler(BaseHandler):
+    def get(self, *args, **kw):
+        self.write_object(self.model)
+
+
 def raise_usage_error(*args, **kw):
     """
     Function that raises a generic :class:`CollectionUsageError`. For use in
@@ -40,6 +46,52 @@ def raise_dummy_error(*args, **kw):
     raise DummyError("You pushed the red button")
 
 
+class TestJoinPaths(TestCase):
+    def test_none(self):
+        self.assertEqual(join_paths(""), "")
+
+    def test_leading_slash(self):
+        self.assertEqual(join_paths("/foo", "bar", "baz"), "/foo/bar/baz")
+
+    def test_trailing_slash(self):
+        self.assertEqual(join_paths("foo", "bar", "baz/"), "foo/bar/baz/")
+
+    def test_leading_and_trailing_slash(self):
+        self.assertEqual(join_paths("/foo", "bar", "baz/"), "/foo/bar/baz/")
+
+    def test_no_slash(self):
+        self.assertEqual(join_paths("foo"), "foo")
+
+    def test_standalone_slash(self):
+        self.assertEqual(join_paths("/"), "/")
+
+    def test_empty_paths(self):
+        self.assertEqual(
+            join_paths("", "foo", "", "bar", "", "baz", ""),
+            "foo/bar/baz")
+
+
+class TestDuplicates(TestCase):
+    def test_none(self):
+        self.assertEqual(duplicates([1, 2, 3]), set())
+
+    def test_duplicates(self):
+        self.assertEqual(duplicates([1, 2, 1, 3, 2, 1]), set([1, 2]))
+
+
+class TestParseRouteVars(TestCase):
+    def test_no_variables(self):
+        self.assertEqual(parse_route_vars("/foo/bar"), [])
+
+    def test_one_variable(self):
+        self.assertEqual(parse_route_vars("/:foo/bar"), ["foo"])
+
+    def test_two_variables(self):
+        self.assertEqual(
+            parse_route_vars("/:foo/bar/:baz"),
+            ["foo", "baz"])
+
+
 class TestCreateUrlspecRegex(TestCase):
     def test_no_variables(self):
         self.assertEqual(create_urlspec_regex("/foo/bar"), "/foo/bar")
@@ -63,10 +115,17 @@ class TestCreateUrlspecRegex(TestCase):
     def test_standalone_slash(self):
         self.assertEqual(create_urlspec_regex("/"), "/")
 
+    def test_duplicate_vars(self):
+        self.assertRaises(
+            RouteParseError,
+            create_urlspec_regex,
+            "/:foo/:bar/baz/:bar/:foo")
+
 
 class TestBaseHandler(TestCase):
     def setUp(self):
-        self.handler_helper = HandlerHelper(BaseHandler)
+        self.handler_helper = HandlerHelper(
+            BaseHandler, {"model_factory": lambda x: None})
 
     def assert_writes(self, writes, expected_objects):
         lines = "".join(writes).rstrip("\n").split("\n")
@@ -140,6 +199,73 @@ class TestBaseHandler(TestCase):
             {"id": "obj2"},
         ])
 
+    def test_mk_urlspec(self):
+        class DummyHandler(BaseHandler):
+            route_suffix = '/baz'
+
+        model_factory = lambda x: None
+        urlspec = DummyHandler.mk_urlspec('/bar/', model_factory, 'foo/')
+
+        self.assertEqual(urlspec.handler_class, DummyHandler)
+        self.assertEqual(urlspec.kwargs, {"model_factory": model_factory})
+        self.assertEqual(urlspec.regex.pattern, 'foo/bar/baz$')
+
+    def test_route_var_attrs(self):
+        class DummyHandler(BaseHandler):
+            route_suffix = '/:foo/:bar'
+
+        self.handler_helper = HandlerHelper(
+            DummyHandler, {"model_factory": lambda x: None})
+
+    def test_initialize(self):
+        model = {}
+        helper = HandlerHelper(BaseHandler, {"model_factory": lambda _: model})
+        handler = helper.mk_handler()
+        self.assertEqual(handler.model_factory(handler), model)
+
+    def test_prepare_model(self):
+        model = {}
+        helper = HandlerHelper(BaseHandler, {"model_factory": lambda _: model})
+        handler = helper.mk_handler()
+        handler.prepare()
+        self.assertEqual(handler.model, model)
+
+    def test_prepare_model_alias(self):
+        class DummyHandler(BaseHandler):
+            model_alias = "foo"
+
+        model = {}
+        helper = HandlerHelper(
+            DummyHandler, {"model_factory": lambda _: model})
+
+        handler = helper.mk_handler()
+        handler.prepare()
+        self.assertEqual(handler.foo, model)
+
+    def test_prepare_no_model_alias(self):
+        class DummyHandler(BaseHandler):
+            model_alias = None
+
+        helper = HandlerHelper(DummyHandler, {"model_factory": lambda _: None})
+        handler = helper.mk_handler()
+
+        # errors if BaseHandler doesn't know how to handle no model alias
+        handler.prepare()
+
+    def test_prepare_route_var_attrs(self):
+        class DummyHandler(BaseHandler):
+            route_suffix = '/:foo/:baz'
+
+        helper = HandlerHelper(DummyHandler, {"model_factory": lambda _: None})
+        handler = helper.mk_handler()
+        handler.path_kwargs = {
+            "foo": "bar",
+            "baz": "quux"
+        }
+        handler.prepare()
+        self.assertEqual(handler.foo, "bar")
+        self.assertEqual(handler.baz, "quux")
+
 
 class BaseHandlerTestCase(TestCase):
     @inlineCallbacks
@@ -161,31 +287,13 @@ class TestCollectionHandler(BaseHandlerTestCase):
             "obj2": {"id": "obj2"},
         }
         self.collection = InMemoryCollection(self.collection_data)
-        self.collection_factory = lambda req: self.collection
+        self.model_factory = lambda req: self.collection
         self.handler_helper = HandlerHelper(
             CollectionHandler,
-            handler_kwargs={'collection_factory': self.collection_factory})
+            handler_kwargs={'model_factory': self.model_factory})
         self.app_helper = AppHelper(
             urlspec=CollectionHandler.mk_urlspec(
-                '/root', self.collection_factory))
-
-    def test_mk_urlspec(self):
-        urlspec = CollectionHandler.mk_urlspec(
-            '/root', self.collection_factory)
-        self.assertEqual(urlspec.handler_class, CollectionHandler)
-        self.assertEqual(urlspec.kwargs, {
-            "collection_factory": self.collection_factory,
-        })
-        self.assertEqual(urlspec.regex.pattern, '/root/$')
-
-    def test_initialize(self):
-        handler = self.handler_helper.mk_handler()
-        self.assertEqual(handler.collection_factory(handler), self.collection)
-
-    def test_prepare(self):
-        handler = self.handler_helper.mk_handler()
-        handler.prepare()
-        self.assertEqual(handler.collection, self.collection)
+                '/root', self.model_factory))
 
     @inlineCallbacks
     def test_get_stream(self):
@@ -268,6 +376,16 @@ class TestCollectionHandler(BaseHandlerTestCase):
         [f] = self.flushLoggedErrors(DummyError)
         self.assertEqual(str(f.value), "You pushed the red button")
 
+    @inlineCallbacks
+    def test_post_bad_json(self):
+        """
+        This test makes sure that if invalid json is received, the ValueError
+        is correctly caught and returned as an HTTP 400 error.
+        """
+        data = yield self.app_helper.post('/root/', data='{', parser='json')
+        self.assertEqual(data.get(u'status_code'), 400)
+        self.assertEqual(data.get(u'reason').find('Invalid JSON: '), 0)
+
 
 class TestElementHandler(BaseHandlerTestCase):
     def setUp(self):
@@ -276,33 +394,13 @@ class TestElementHandler(BaseHandlerTestCase):
             "obj2": {"id": "obj2"},
         }
         self.collection = InMemoryCollection(self.collection_data)
-        self.collection_factory = lambda req: self.collection
+        self.model_factory = lambda req: self.collection
         self.handler_helper = HandlerHelper(
             ElementHandler,
-            handler_kwargs={'collection_factory': self.collection_factory})
+            handler_kwargs={'model_factory': self.model_factory})
         self.app_helper = AppHelper(
             urlspec=ElementHandler.mk_urlspec(
-                '/root', self.collection_factory))
-
-    def test_mk_urlspec(self):
-        urlspec = ElementHandler.mk_urlspec(
-            '/root', self.collection_factory)
-        self.assertEqual(urlspec.handler_class, ElementHandler)
-        self.assertEqual(urlspec.kwargs, {
-            "collection_factory": self.collection_factory,
-        })
-        self.assertEqual(urlspec.regex.pattern, '/root/(?P<elem_id>[^/]*)$')
-
-    def test_initialize(self):
-        handler = self.handler_helper.mk_handler()
-        self.assertEqual(handler.collection_factory(handler), self.collection)
-
-    def test_prepare(self):
-        handler = self.handler_helper.mk_handler()
-        handler.path_kwargs = {"elem_id": "id-1"}
-        handler.prepare()
-        self.assertEqual(handler.collection, self.collection)
-        self.assertEqual(handler.elem_id, "id-1")
+                '/root', self.model_factory))
 
     @inlineCallbacks
     def test_get(self):
@@ -369,6 +467,20 @@ class TestElementHandler(BaseHandlerTestCase):
         [f] = self.flushLoggedErrors(DummyError)
         self.assertEqual(str(f.value), "You pushed the red button")
 
+    @inlineCallbacks
+    def test_put_bad_json(self):
+        """
+        This test makes sure that if invalid json is received, the ValueError
+        is correctly caught and returned as an HTTP 400 error.
+        """
+        self.assertEqual(self.collection_data["obj2"], {"id": "obj2"})
+        data = yield self.app_helper.put(
+            '/root/obj2',
+            data='{',
+            parser='json')
+        self.assertEqual(data.get(u'status_code'), 400)
+        self.assertEqual(data.get(u'reason').find('Invalid JSON: '), 0)
+
     @inlineCallbacks
     def test_delete(self):
         self.assertTrue("obj1" in self.collection_data)
@@ -408,12 +520,17 @@ class TestApiApplication(TestCase):
         self.collection_helper = HandlerHelper(
             CollectionHandler,
             handler_kwargs={
-                "collection_factory": self.uncallable_collection_factory,
+                "model_factory": self.uncallable_model_factory,
             })
         self.element_helper = HandlerHelper(
             ElementHandler,
             handler_kwargs={
-                "collection_factory": self.uncallable_collection_factory,
+                "model_factory": self.uncallable_model_factory,
+            })
+        self.dummy_helper = HandlerHelper(
+            DummyHandler,
+            handler_kwargs={
+                "model_factory": self.uncallable_model_factory,
             })
 
     @inlineCallbacks
@@ -438,23 +555,29 @@ class TestApiApplication(TestCase):
         collection = InMemoryCollection(data)
         return lambda _req: collection
 
-    def get_app_helper(self, collections=ApiApplication.collections,
-                       preprocessor=(
-                           ApiApplication.collection_factory_preprocessor),
-                       config=None, extra_settings=None):
+    def get_app_helper(self,
+                       collections=ApiApplication.collections,
+                       models=ApiApplication.models,
+                       preprocessor=ApiApplication.factory_preprocessor,
+                       config=None,
+                       extra_settings=None):
         class MyApiApplication(ApiApplication):
             pass
 
         MyApiApplication.collections = collections
+        MyApiApplication.models = models
+
         if callable(preprocessor):
             preprocessor = staticmethod(preprocessor)
-        MyApiApplication.collection_factory_preprocessor = preprocessor
+
+        MyApiApplication.factory_preprocessor = preprocessor
         if extra_settings is None:
             extra_settings = {}
+
         return AppHelper(MyApiApplication(config, **extra_settings))
 
     @inlineCallbacks
-    def test_process_request_health_check(self):
+    def test_process_health_request_check(self):
         app_helper = self.get_app_helper()
         result = yield app_helper.request('GET', '/health/')
         content = yield result.content()
@@ -462,11 +585,11 @@ class TestApiApplication(TestCase):
         self.assertEqual(content, "OK")
 
     @inlineCallbacks
-    def test_process_request_default_preprocessor(self):
+    def test_process_collection_request_default_preprocessor(self):
         collection_data = {'foo': {'id': 'foo'}}
-        collection_factory = self.get_collection_factory(collection_data)
+        model_factory = self.get_collection_factory(collection_data)
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),))
+            collections=(('/:owner_id/store', model_factory),))
         result = yield app_helper.request(
             'GET', '/foo/store/?stream=true',
             headers={"X-Owner-ID": "owner-1"})
@@ -474,134 +597,249 @@ class TestApiApplication(TestCase):
         self.assertEqual(json.loads(content), collection_data['foo'])
 
     @inlineCallbacks
-    def test_process_request_no_preprocessor(self):
+    def test_process_collection_request_no_preprocessor(self):
         collection_data = {'foo': {'id': 'foo'}}
-        collection_factory = self.get_collection_factory(collection_data)
+        model_factory = self.get_collection_factory(collection_data)
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             preprocessor=None)
         result = yield app_helper.request('GET', '/foo/store/?stream=true')
         content = yield result.content()
         self.assertEqual(json.loads(content), collection_data['foo'])
 
     @inlineCallbacks
-    def test_process_request_async_preprocessor(self):
+    def test_process_collection_request_async_preprocessor(self):
         collection_data = {'foo': {'id': 'foo'}}
-        collection_factory = self.get_collection_factory(collection_data)
+        model_factory = self.get_collection_factory(collection_data)
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             preprocessor=lambda handler: succeed("owner-1"))
         result = yield app_helper.request('GET', '/foo/store/?stream=true')
         content = yield result.content()
         self.assertEqual(json.loads(content), collection_data['foo'])
 
     @inlineCallbacks
-    def test_process_request_bouncer_preprocessor(self):
+    def test_process_collection_request_bouncer_preprocessor(self):
         collection_data = {'foo': {'id': 'foo'}}
-        collection_factory = self.get_collection_factory(collection_data)
+        model_factory = self.get_collection_factory(collection_data)
         auth_server = yield self.start_fake_auth_server("owner-1")
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             preprocessor=owner_from_oauth2_bouncer(auth_server.url))
         result = yield app_helper.request('GET', '/foo/store/?stream=true')
         content = yield result.content()
         self.assertEqual(json.loads(content), collection_data['foo'])
 
-    def uncallable_collection_factory(self, *args, **kw):
+    @inlineCallbacks
+    def test_process_model_request_default_preprocessor(self):
+        model = {'foo': 'bar'}
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, lambda _: model),))
+
+        result = yield app_helper.request(
+            'GET', '/foo',
+            headers={"X-Owner-ID": "owner-1"})
+        content = yield result.content()
+        self.assertEqual(json.loads(content), model)
+
+    @inlineCallbacks
+    def test_process_model_request_no_preprocessor(self):
+        model = {'foo': 'bar'}
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, lambda _: model),),
+            preprocessor=None)
+
+        result = yield app_helper.request('GET', '/foo')
+        content = yield result.content()
+        self.assertEqual(json.loads(content), model)
+
+    @inlineCallbacks
+    def test_process_model_request_async_preprocessor(self):
+        model = {'foo': 'bar'}
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, lambda _: model),),
+            preprocessor=lambda handler: succeed("owner-1"))
+
+        result = yield app_helper.request('GET', '/foo')
+        content = yield result.content()
+        self.assertEqual(json.loads(content), model)
+
+    @inlineCallbacks
+    def test_process_model_request_bouncer_preprocessor(self):
+        model = {'foo': 'bar'}
+        auth_server = yield self.start_fake_auth_server("owner-1")
+
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, lambda _: model),),
+            preprocessor=owner_from_oauth2_bouncer(auth_server.url))
+
+        result = yield app_helper.request('GET', '/foo')
+        content = yield result.content()
+        self.assertEqual(json.loads(content), model)
+
+    def uncallable_model_factory(self, *args, **kw):
         """
-        A collection_factory for use in tests that need one but should never
+        A model_factory for use in tests that need one but should never
         call it.
         """
-        raise Exception("This collection_factory should never be called")
+        raise Exception("This model_factory should never be called")
 
     def test_build_routes_no_preprocesor(self):
-        collection_factory = self.uncallable_collection_factory
+        model_factory = self.uncallable_model_factory
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
+            models=(('/foo', DummyHandler, model_factory),),
             preprocessor=None)
         routes = app_helper.app.handlers[0][1]
-        [_health_route, collection_route, elem_route] = routes
+        [_health_route, collection_route, elem_route, model_route] = routes
         self.assertEqual(collection_route.handler_class, CollectionHandler)
         self.assertEqual(collection_route.regex.pattern,
                          "/(?P<owner_id>[^/]*)/store/$")
         self.assertEqual(collection_route.kwargs, {
-            "collection_factory": collection_factory,
+            "model_factory": model_factory,
         })
         self.assertEqual(elem_route.handler_class, ElementHandler)
         self.assertEqual(elem_route.regex.pattern,
                          "/(?P<owner_id>[^/]*)/store/(?P<elem_id>[^/]*)$")
         self.assertEqual(elem_route.kwargs, {
-            "collection_factory": collection_factory,
+            "model_factory": model_factory,
+        })
+        self.assertEqual(model_route.handler_class, DummyHandler)
+        self.assertEqual(model_route.regex.pattern, "/foo$")
+        self.assertEqual(model_route.kwargs, {
+            "model_factory": model_factory,
         })
 
     @inlineCallbacks
-    def assert_handlers_get_owner(self, app, collection_name, **handler_kw):
-        [_health_route, collection_route, elem_route] = app.handlers[0][1]
+    def assert_collection_handlers_get_owner(self, app, collection_name,
+                                              **handler_kw):
+        [_health_route, collection_route, elem_route] = app.handlers[0][1][:3]
 
         handler = self.collection_helper.mk_handler(**handler_kw)
-        collection_factory = collection_route.kwargs["collection_factory"]
-        owner = yield collection_factory(handler)
+        model_factory = collection_route.kwargs["model_factory"]
+        owner = yield model_factory(handler)
         self.assertEqual(owner, collection_name)
 
         handler = self.element_helper.mk_handler(**handler_kw)
-        collection_factory = elem_route.kwargs["collection_factory"]
-        owner = yield collection_factory(handler)
+        model_factory = elem_route.kwargs["model_factory"]
+        owner = yield model_factory(handler)
         self.assertEqual(owner, collection_name)
 
     @inlineCallbacks
-    def test_build_routes_with_default_preprocessor(self):
-        collection_factory = lambda owner_id: "collection-%s" % owner_id
-        app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),))
+    def assert_model_handlers_get_owner(self, app, model_name, **handler_kw):
+        for model_route in app.handlers[0][1][3:]:
+            handler = self.model_helper.mk_handler(**handler_kw)
+            model_factory = model_route.kwargs["model_factory"]
+            owner = yield model_factory(handler)
+            self.assertEqual(owner, model_name)
 
-        yield self.assert_handlers_get_owner(
+    @inlineCallbacks
+    def test_build_collection_routes_with_default_preprocessor(self):
+        model_factory = lambda owner_id: "collection-%s" % owner_id
+        app_helper = self.get_app_helper(
+            collections=(('/:owner_id/store', model_factory),))
+
+        yield self.assert_collection_handlers_get_owner(
             app_helper.app, "collection-owner-1",
             headers={"X-Owner-ID": "owner-1"})
 
     @inlineCallbacks
-    def test_build_routes_with_header_preprocessor(self):
-        collection_factory = lambda owner_id: "collection-%s" % owner_id
+    def test_build_collection_routes_with_header_preprocessor(self):
+        model_factory = lambda owner_id: "collection-%s" % owner_id
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             preprocessor=owner_from_header("X-Foo-ID"))
 
-        yield self.assert_handlers_get_owner(
+        yield self.assert_collection_handlers_get_owner(
             app_helper.app, "collection-owner-1",
             headers={"X-Foo-ID": "owner-1"})
 
     @inlineCallbacks
-    def test_build_routes_with_path_kwargs_preprocessor(self):
-        collection_factory = lambda owner_id: "collection-%s" % owner_id
+    def test_build_collection_routes_with_path_kwargs_preprocessor(self):
+        model_factory = lambda owner_id: "collection-%s" % owner_id
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             preprocessor=owner_from_path_kwarg("owner_id"))
 
-        yield self.assert_handlers_get_owner(
+        yield self.assert_collection_handlers_get_owner(
             app_helper.app, "collection-owner-1",
             path_kwargs={"owner_id": "owner-1"})
 
     @inlineCallbacks
-    def test_build_routes_with_async_preprocessor(self):
-        collection_factory = lambda owner_id: "collection-%s" % owner_id
+    def test_build_collection_routes_with_async_preprocessor(self):
+        model_factory = lambda owner_id: "collection-%s" % owner_id
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             preprocessor=lambda handler: succeed("owner-1"))
 
-        yield self.assert_handlers_get_owner(
+        yield self.assert_collection_handlers_get_owner(
             app_helper.app, "collection-owner-1")
 
     @inlineCallbacks
-    def test_build_routes_with_bouncer_preprocessor(self):
+    def test_build_collection_routes_with_bouncer_preprocessor(self):
         auth_server = yield self.start_fake_auth_server("owner-1")
-        collection_factory = lambda owner_id: "collection-%s" % owner_id
+        model_factory = lambda owner_id: "collection-%s" % owner_id
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             preprocessor=owner_from_oauth2_bouncer(auth_server.url))
 
-        yield self.assert_handlers_get_owner(
+        yield self.assert_collection_handlers_get_owner(
             app_helper.app, "collection-owner-1",
             headers={"Authorization": "Bearer token"})
 
+    @inlineCallbacks
+    def test_build_model_routes_with_default_preprocessor(self):
+        model_factory = lambda owner_id: "model-%s" % owner_id
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, model_factory),))
+
+        yield self.assert_model_handlers_get_owner(
+            app_helper.app, "model-owner-1",
+            headers={"X-Owner-ID": "owner-1"})
+
+    @inlineCallbacks
+    def test_build_model_routes_with_header_preprocessor(self):
+        model_factory = lambda owner_id: "model-%s" % owner_id
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, model_factory),),
+            preprocessor=owner_from_header("X-Foo-ID"))
+
+        yield self.assert_model_handlers_get_owner(
+            app_helper.app, "model-owner-1",
+            headers={"X-Foo-ID": "owner-1"})
+
+    @inlineCallbacks
+    def test_build_model_routes_with_path_kwargs_preprocessor(self):
+        model_factory = lambda owner_id: "model-%s" % owner_id
+        app_helper = self.get_app_helper(
+            models=(('/:owner_id/foo', DummyHandler, model_factory),),
+            preprocessor=owner_from_path_kwarg("owner_id"))
+
+        yield self.assert_model_handlers_get_owner(
+            app_helper.app, "model-owner-1")
+
+    @inlineCallbacks
+    def test_build_model_routes_with_async_preprocessor(self):
+        model_factory = lambda owner_id: "model-%s" % owner_id
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, model_factory),),
+            preprocessor=lambda handler: succeed("owner-1"))
+
+        yield self.assert_model_handlers_get_owner(
+            app_helper.app, "model-owner-1")
+
+    @inlineCallbacks
+    def test_build_model_routes_with_bouncer_preprocessor(self):
+        auth_server = yield self.start_fake_auth_server("owner-1")
+        model_factory = lambda owner_id: "model-%s" % owner_id
+        app_helper = self.get_app_helper(
+            models=(('/foo', DummyHandler, model_factory),),
+            preprocessor=owner_from_oauth2_bouncer(auth_server.url))
+
+        yield self.assert_model_handlers_get_owner(
+            app_helper.app, "model-owner-1",
+            headers={"Authorization": "Bearer token"})
+
     def test_get_config_settings_None(self):
         app = ApiApplication()
         self.assertEqual(app.get_config_settings(), {})
@@ -653,14 +891,14 @@ class TestApiApplication(TestCase):
         # that we don't have the default preprocessor.
         app = ApiApplication(tempfile)
         self.assertNotEqual(
-            app.collection_factory_preprocessor,
-            ApiApplication.collection_factory_preprocessor)
+            app.factory_preprocessor,
+            ApiApplication.factory_preprocessor)
 
         # With no config specified, we should have the default preprocessor.
         app = ApiApplication()
         self.assertEqual(
-            app.collection_factory_preprocessor,
-            ApiApplication.collection_factory_preprocessor)
+            app.factory_preprocessor,
+            ApiApplication.factory_preprocessor)
 
     def test_configure_url_path_prefix(self):
         config_dict = {'url_path_prefix': '/foo/bar'}
@@ -670,12 +908,13 @@ class TestApiApplication(TestCase):
         with open(tempfile, 'wb') as fp:
             yaml.safe_dump(config_dict, fp)
 
-        collection_factory = self.get_collection_factory({})
+        model_factory = self.get_collection_factory({})
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
+            models=(('/baz', DummyHandler, self.uncallable_model_factory),),
             config=tempfile)
 
-        [health, collection, element] = app_helper.app.handlers[0][1]
+        [health, collection, element, model] = app_helper.app.handlers[0][1]
         # No prefix for the health URL path.
         self.assertEqual(health.regex.pattern, r'/health/$')
         # We should have the prefix on the front of our other URL paths.
@@ -685,13 +924,17 @@ class TestApiApplication(TestCase):
         self.assertEqual(
             element.regex.pattern,
             r'/foo/bar/(?P<owner_id>[^/]*)/store/(?P<elem_id>[^/]*)$')
+        self.assertEqual(
+            model.regex.pattern,
+            r'/foo/bar/baz$')
 
     def test_no_url_path_prefix(self):
-        collection_factory = self.get_collection_factory({})
+        model_factory = self.get_collection_factory({})
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),))
+            collections=(('/:owner_id/store', model_factory),),
+            models=(('/baz', DummyHandler, self.uncallable_model_factory),))
 
-        [health, collection, element] = app_helper.app.handlers[0][1]
+        [health, collection, element, model] = app_helper.app.handlers[0][1]
 
         self.assertEqual(health.regex.pattern, r'/health/$')
         self.assertEqual(
@@ -700,13 +943,16 @@ class TestApiApplication(TestCase):
         self.assertEqual(
             element.regex.pattern,
             r'/(?P<owner_id>[^/]*)/store/(?P<elem_id>[^/]*)$')
+        self.assertEqual(
+            model.regex.pattern,
+            r'/baz$')
 
     @inlineCallbacks
     def test_handler_log_suppression(self):
         handler_logs = []
-        collection_factory = self.get_collection_factory({})
+        model_factory = self.get_collection_factory({})
         app_helper = self.get_app_helper(
-            collections=(('/:owner_id/store', collection_factory),),
+            collections=(('/:owner_id/store', model_factory),),
             extra_settings={'log_function': handler_logs.append})
 
         # The collection handler should use default logging behaviour.
@@ -726,7 +972,7 @@ class TestAuthHandlers(TestCase):
         self.dummy_helper = HandlerHelper(
             CollectionHandler,
             handler_kwargs={
-                "collection_factory": self.uncallable_collection_factory,
+                "model_factory": self.uncallable_model_factory,
             })
 
     @inlineCallbacks
@@ -752,12 +998,12 @@ class TestAuthHandlers(TestCase):
         self.add_cleanup(fake_server.stop)
         returnValue(fake_server)
 
-    def uncallable_collection_factory(self, *args, **kw):
+    def uncallable_model_factory(self, *args, **kw):
         """
-        A collection_factory for use in tests that need one but should never
+        A model_factory for use in tests that need one but should never
         call it.
         """
-        raise Exception("This collection_factory should never be called")
+        raise Exception("This model_factory should never be called")
 
     @inlineCallbacks
     def test_fake_auth_server_rejects_post(self):
